########################################################################
#
# File Name:            dump.py
#
# Documentation:        http://docs.4suite.org/4RDF/dump.py.html
#
"""
Command-line invokation of the 4RDF processor
WWW: http://4suite.org/4RDF        e-mail: support@4suite.org

Copyright (c) 2001 Fourthought, Inc., USA.   All Rights Reserved.
See  http://4suite.org/COPYRIGHT  for license and copyright information
"""

import sys
from Ft import __version__
from Ft.Rdf import Model
from Ft.Rdf.Serializers.Dom import Serializer

def Run(options,args):
    source = args['source-uri']

    if dump:
        driver = "Memory"
    elif not driver:
        raise SystemExit("No driver specified")

    from Ft.Rdf.Drivers import Memory

    # Memory driver simply creates a new db anyway
    db = Memory.GetDb(dbname)
    db.begin()
    model = Model.Model(db)

    from Ft.Xml import Domlette
    reader = Domlette.NonvalidatingReader()
    if source == '-':
        doc = reader.parseStream(sys.stdin)
    else:
        doc = reader.parseUri(source)

    serializer = Serializer()
    serializer.deserialize(model, doc, '')

    if dump:
        stmts = model.complete(None, None, None)
        print 'The following is a list of remaining tuples, each in the form "subject, predicate, object".'
        print "["
        for stmt in stmts:
            print '("%s", "%s", "%s"),' % (stmt.subject,
                                           stmt.predicate,
                                           stmt.object)
        print "]"

    db.commit()
    return






from Ft.Lib.CommandLine import Options, CommandLineApp, Arguments, Command

class RdfCommandLineApp(CommandLineApp.CommandLineApp):

    def __init__(self):
        CommandLineApp.CommandLineApp.__init__(
            self,
            '4rdf',
            '4RDF version %s' % __version__,
            __doc__,
            [],
            ourOptions=Options.Options([Options.Option('d',
                                                       'dump',
                                                       'Dump the resulting triples'\
                                                       'to standard output (no presistent'\
                                                       'changes are made)',
                                                       ),
                                        Options.Option(None,
                                                       'driver=',
                                                       'Use the given 4RDF backend driver.',
                                                       ),
                                        Options.Option(None,
                                                       'dbName=',
                                                       'Update the RDF model database'\
                                                       'with the given name', 
                                                       ),
                                        ]),
            enableShowCommands=0
            )

        self.function = Run
        self.arguments = [Arguments.RequiredArgument('source-uri',
                                                     'The URI of the source RDF document.  Note: if you use "-" as the name'\
                                                     'of the source document, the source will  instead be read from standard input.',
                                                     str),
                          ]


    def validate_arguments(self, args):
        return Command.Command.validate_arguments(self, args)


